
package Ham::RViesti::Client;

=head1 NAME

Ham::RViesti::Client - A client for accessing the R.Viesti backend service

=head1 ABSTRACT

This module provides access to the RATS R.Viesti POCSAG short messaging
system server's API. It can be used to implement message sourcing systems
which send short messages to users and user groups within the R.Viesti
infrastructure, and POCSAG transmitter clients which fetch messages from
the servers and drive the radio transmitters.

=head1 DESCRIPTION

Unless a debugging mode is enabled, all errors and warnings are reported
through the API (as opposed to printing on STDERR or STDOUT), so that
they can be reported nicely on the user interface of an application.

=cut

use Data::Dumper;

# API responses are in JSON
use JSON;

# used to access the API
use LWP;
use LWP::UserAgent;
use HTTP::Request::Common;
use URI::Escape qw(uri_escape);

# used for API message authentication
use Digest::HMAC_SHA1 qw(hmac_sha1_hex);

our $VERSION = '1.00';

=head1 OBJECT INTERFACE

=over

=item new(config)

Returns a new instance of the R.Viesti perl client.

  my $kauha = new Ham::RViesti::Client(
    'app' => 'myappname',
    'apikey' => '42.vh7fy3h', # your API key
    'apisecret' => 'jca973hr3hf', # your API secret
    'debug' => 0, # set to 1 for debug log on stderr
  );

=back 

=cut

sub new 
{
	my $class = shift;
	my $self = bless { @_ }, $class;
	
	$self->{'initialized'} = 0;
	$self->{'name'} = 'Ham::RViesti::Client';
	$self->{'version'} = $VERSION;
	
	# store config
	my %h = @_;
	$self->{'config'} = \%h;
	#print "settings: " . Dumper(\%h);
	
	$self->{'debug'} = ( $self->{'config'}->{'debug'} );
	
	$self->_debug('initializing');
	
	$self->_clear_errors();
	
	# validate settings
	foreach my $k ('app', 'apikey', 'apisecret') {
		if (!defined $h{$k}) {
			return $self->_critical("Mandatory config setting '$k' not set!");
		}
	}
	
	if (!defined $h{'service'}) {
		$h{'service'} = [ 'rv-api-1.rats.fi:8042', 'rv-api-2.rats.fi:8042' ];
	}
	
	if (defined $h{'mode'} && $h{'mode'} == 'each') {
		$self->{'mode'} = 'each';
	} else {
		$self->{'mode'} = 'any';
	}
	
	# initialize HTTP client
	$ua = LWP::UserAgent->new;
	$ua->agent(
		agent => $self->{'name'} . '/' . $self->{'version'},
		timeout => 30,
		max_redirect => 0,
	);
	
	$self->{'ua'} = $ua;
	
	# initialize JSON parser
	#$json = new JSON::XS;
	$json = new JSON;
	
	if (!$json) {
		return $self->_critical("JSON loading failed");
	}
	
	#$json->latin1(0);
	#$json->ascii(1);
	#$json->utf8(0);
	
	$self->{'json'} = $json;
	
	# next server
	$self->{'next_server'} = 0;
	$self->{'max_server'} = $#{ $self->{'config'}->{'service'} };
	$self->_next_server();
	
	return $self;
}

# report a critical error

sub _critical($$)
{
	my($self, $msg) = @_;
	
	warn $self->{'name'} . " - " . $msg . "\n";
	
	$self->{'last_err_code'} = 'CRITICAL';
	$self->{'last_err_msg'} = $msg;
	
	return;
}

# report an error

sub _error($$$)
{
	my($self, $code, $msg) = @_;
	
	if ($self->{'debug'}) {
		warn $self->{'name'} . " ERROR $code: $msg\n";
	}
	
	$self->{'last_err_code'} = $code;
	$self->{'last_err_msg'} = $msg;
	
	return 0;
}

=over

=item get_error()

Returns the error code and message string for the most recent error
experienced by the module. Usage:

  my($code, $msg) = $client->get_error();

=back

=cut

sub get_error($)
{
	return ($self->{'last_err_code'}, $self->{'last_err_msg'});
}

# clear the error flags

sub _clear_errors($)
{
	$self->{'last_err_code'} = 'ok';
	$self->{'last_err_msg'} = 'no error reported';
}

# report a debug log

sub _debug($$)
{
	my($self, $msg) = @_;
	
	return if (!$self->{'debug'});
	
	warn $self->{'name'} . " DEBUG $msg\n";
}

=over

=item message_submit()

  $kauha->message_submit(
    'msgid' => $unique_message_id_string,
    'to' => $callsign,
    'message' => $message
  );

Submits a message to a callsign. The message ID must be an unique ASCII string shorter than 256 bytes.
The 'to' field specifies a callsign, or the name of a distribution list like '@OHDX', or the RIC
number code directly to transmit to.

Transmitting directly to RIC codes is good for testing, but using the R.Viesti infrastructure for
mapping callsigns to recipient numbers is recommended, so that the recipient can easily configure
a new pager in a single place.

=back

=cut

sub message_submit($@)
{
	my($self, @args) = @_;
	
	$self->_clear_errors();
	
	# validate options
	my %options = @args;
	
	foreach my $k ('msgid', 'to', 'message') {
		if (!defined $options{$k}) {
			return $self->_error('err_invalid_arg', "Missing argument: $k");
		}
	}
	
	my $response = $self->_api_call('submit',
		'cli' => $self->{'name'},
		'cliver' => $VERSION,
		'app' => $self->{'config'}->{'app'},
		'msgid' => $options{'msgid'},
		'to' => $options{'to'},
		'message' => $options{'message'}
	);
	
	return $response if (!$response);
	
	if (!defined $response->{'queued'}) {
		return $self->_error('submit_no_queued', "API did not return 'queued' count for message submit");
	}
	
	if ($response->{'queued'} < 1) {
		return $self->_error('submit_not_sent', "Server failed to queue message: "
			. (defined $response->{'description'} ? $response->{'description'} : '(no description)')
			);
	}
	
	return 1;
}

=over

=item tx_fetch()

 my $blob = $kauha->tx_fetch();

Fetches messages from the servers for a transmitter to transmit.

=back

=cut

sub tx_fetch($;@)
{
	my($self, @args) = @_;
	
	$self->_clear_errors();
	
	# validate options
	#my %options = @args;
	
	return $self->_api_call('tx', 
		'foo' => 'bar',
	);
}

# figure out the next server URL to use

sub _next_server($)
{
	my($self) = @_;
	
	my $server = $self->{'config'}->{'service'}->[ $self->{'next_server'} ];
	my $url = "http://$server/api/";
	
	$self->_debug('switching to next server: ' . $server);
	
	# rotate
	$self->{'next_server'} += 1;
	$self->{'next_server'} = 0 if ($self->{'next_server'} > $self->{'max_server'});
	
	$self->{'this_server'} = $url;
	
	return $url;
}

# the last server call failed

sub _server_fail($)
{
	my($self) = @_;
	
	# if we're in "talk to any server that answers first" mode,
	# switch to the next server since this one didn't work
	if ($self->{'mode'} eq 'any') {
		$self->_debug("server failed, switching");
		$self->_next_server();
	}
}

sub _server_ok($)
{
	my($self) = @_;
	
	# if we're in "talk to each server sequentially" mode,
	# switch to the next server
	if ($self->{'mode'} eq 'each') {
		$self->_debug("server ok, switching");
		$self->_next_server();
	}
}

# submit an API call

sub _api_call($$@)
{
	my($self, $command, @args) = @_;
	
	$self->_clear_errors();
	my %options = @args;
	
	# put authentication variables in
	$options{'apikey'} = $self->{'config'}->{'apikey'};
	# this is not really required, it just adds entropy
	$options{'apinonce'} = time() . $$ . rand(1000000000);
	
	# generate a message to send - it needs to be sorted,
	# so that the hmac calculation happens in the correct order at
	# both ends!
	my @a;
	my @h;
	foreach my $k (sort keys %options) {
		push @a, "$k=" . uri_escape($options{$k});
		push @h, "$k=$options{$k}";
	}
	
	# calculate HMAC, which will be passed as a HTTP header
	my $s = join('&', @h);
	my $apiauth = hmac_sha1_hex($s, $self->{'config'}->{'apisecret'});
	
	# Ok, then post the message over HTTP
	my $baseurl = $self->{'this_server'};
	my $url = $baseurl . $command;
	
	$self->_debug('performing HTTP API request');
	my $req = HTTP::Request->new(GET => $url . '?' . join('&', @a));
	$req->header('X-API-AUTH' => $apiauth);
	my $res = $self->{'ua'}->simple_request($req);
	
	if ($res->status_line !~ /^(\d+)\s+(.*)$/) {
		$self->_server_fail();
		return $self->_error('err_server',
			"HTTP status line could not be parsed for " . $srv->{'url'} . ": " . $res->status_line);
	}
	
	my($status_code, $status_message) = ($1, $2);
	if ($status_code ne 200) {
		$self->_server_fail();
		return $self->_error('err_server',
			"HTTP error $status_code: $status_message - " . $srv->{'url'});
	}
	
	if ($res->header('Content-Type') !~ /application\/json/) {
		$self->_server_fail();
		return $self->_error('err_server',
			"Server returned wrong content type (should be application/json): " . $res->header('Content-Type') . " - " . $srv->{'url'});
	}
	
	$self->_debug('API HTTP call successful');
	$self->_server_ok();
	return $self->_process_response($url, $res);
}

sub _process_response($$$)
{
	my($self, $url, $res) = @_;
	
	my $blob;
	if ($JSON::VERSION > 2) {
		$blob = $self->{'json'}->decode($res->content);
	} else {
		$blob = $self->{'json'}->jsonToObj($res->content);
	}
	
	if (!$blob) {
		return $self->_error('err_decode', "JSON decoding failed for $url");
	}
	
	if (!defined $blob->{'protocol'}) {
		return $self->_error('err_protocol', "JSON response does not specify protocol");
	}
	
	if ($blob->{'protocol'} ne 'kauha') {
		return $self->_error('err_unknown_protocol', "JSON response specifies unknown protocol: " . $blob->{'protocol'});
	}
	
	$self->_debug('API call response decoded successfully');
	
	if (defined $blob->{'result'}) {
		$self->_debug('API call response contains result: ' . $blob->{'result'});
	}
	
	if (!defined $blob->{'result'}) {
		return $self->_error('err_no_result', 'JSON response does not contain result code');
	}
	
	if ($blob->{'result'} ne 'ok') {
		return $self->_error('api_returned_err', 'API returned result "' . $blob->{'result'} . '": ' . 
			(defined $blob->{'description'} ? $blob->{'description'} : '(no description)')
			);
	}
	
	return $blob;
}

1;

